@@ -29,7 +29,7 @@
 #include <QtQuick3DRuntimeRender/private/qssgrhicontext_p.h>
 #include <QtQuick3DRuntimeRender/private/qssgperframeallocator_p.h>
 #include <QtQuick3DRuntimeRender/private/qssgshadermapkey_p.h>
-#include <QtQuick3DRuntimeRender/private/qssglightmapper_p.h>
+#include <QtQuick3DRuntimeRender/private/qssglightmapbaker_p.h>
 #include <ssg/qssgrenderextensions.h>
 
 #include <QtQuick3DUtils/private/qssgrenderbasetypes_p.h>
@@ -38,12 +38,13 @@
 #include <unordered_map>
 
 #include "qssgrenderpass_p.h"
-
-#define QSSG_RENDER_MINIMUM_RENDER_OPACITY .01f
+#include "qssgrenderdata_p.h"
 
 QT_BEGIN_NAMESPACE
 
-struct QSSGRenderableObject;
+class QSSGRenderableObject;
+
+class QSGRenderer;
 
 enum class QSSGLayerRenderPreparationResultFlag
 {
@@ -67,7 +68,13 @@ enum class QSSGLayerRenderPreparationResultFlag
     RequiresScreenTexture = 1 << 5,
 
     // set together with RequiresScreenTexture when SCREEN_MIP_TEXTURE is used
-    RequiresMipmapsForScreenTexture = 1 << 6
+    RequiresMipmapsForScreenTexture = 1 << 6,
+
+    // Set when material has custom blend mode(not SourceOver)
+    MaterialHasCustomBlendMode = 1 << 7,
+
+    // Set when multisampled depth texture is required
+    RequiresDepthTextureMS = 1 << 8
 };
 
 struct QSSGLayerRenderPreparationResultFlags : public QFlags<QSSGLayerRenderPreparationResultFlag>
@@ -93,6 +100,15 @@ struct QSSGLayerRenderPreparationResultFlags : public QFlags<QSSGLayerRenderPrep
         setFlag(QSSGLayerRenderPreparationResultFlag::RequiresDepthTexture, inValue);
     }
 
+    bool requiresDepthTextureMS() const
+    {
+        return this->operator&(QSSGLayerRenderPreparationResultFlag::RequiresDepthTextureMS);
+    }
+    void setRequiresDepthTextureMS(bool inValue)
+    {
+        setFlag(QSSGLayerRenderPreparationResultFlag::RequiresDepthTextureMS, inValue);
+    }
+
     bool requiresSsaoPass() const { return this->operator&(QSSGLayerRenderPreparationResultFlag::RequiresSsaoPass); }
     void setRequiresSsaoPass(bool inValue)
     {
@@ -125,6 +141,15 @@ struct QSSGLayerRenderPreparationResultFlags : public QFlags<QSSGLayerRenderPrep
     {
         setFlag(QSSGLayerRenderPreparationResultFlag::RequiresMipmapsForScreenTexture, inValue);
     }
+
+    bool hasCustomBlendMode() const
+    {
+        return this->operator&(QSSGLayerRenderPreparationResultFlag::MaterialHasCustomBlendMode);
+    }
+    void setHasCustomBlendMode(bool inValue)
+    {
+        setFlag(QSSGLayerRenderPreparationResultFlag::MaterialHasCustomBlendMode, inValue);
+    }
 };
 
 struct QSSGLayerRenderPreparationResult
@@ -139,7 +164,6 @@ struct QSSGLayerRenderPreparationResult
     bool isNull() const { return !layer; }
     bool isLayerVisible() const;
     QSize textureDimensions() const;
-    QSSGCameraGlobalCalculationResult setupCameraForRender(QSSGRenderCamera &inCamera, float dpr = 1.0f);
 };
 
 struct QSSGDefaultMaterialPreparationResult
@@ -164,6 +188,19 @@ struct QSSGBakedLightingModel
     QVector<QSSGRenderableObjectHandle> renderables;
 };
 
+struct QSSGOITRenderContext
+{
+    QRhiTextureRenderTarget *oitRenderTarget = nullptr;
+    QRhiRenderPassDescriptor *renderPassDescriptor = nullptr;
+    void reset()
+    {
+        delete oitRenderTarget;
+        delete renderPassDescriptor;
+        oitRenderTarget = nullptr;
+        renderPassDescriptor = nullptr;
+    }
+};
+
 class Q_QUICK3DRUNTIMERENDER_EXPORT QSSGLayerRenderData
 {
 public:
@@ -172,6 +209,17 @@ public:
         MAX_TEMPORAL_AA_LEVELS = 2,
     };
 
+    using InstanceTransforms = QSSGGlobalRenderNodeData::InstanceTransforms;
+    using ModelViewProjections = QSSGRenderModelData::ModelViewProjections;
+
+    using QSSGModelsView = QSSGDataView<QSSGRenderModel *>;
+    using QSSGParticlesView = QSSGDataView<QSSGRenderParticles *>;
+    using QSSGItem2DsView = QSSGDataView<QSSGRenderItem2D *>;
+    using QSSGCamerasView = QSSGDataView<QSSGRenderCamera *>;
+    using QSSGLightsView = QSSGDataView<QSSGRenderLight *>;
+    using QSSGReflectionProbesView = QSSGDataView<QSSGRenderReflectionProbe *>;
+    using QSSGNonCategorizedView = QSSGDataView<QSSGRenderNode *>;
+
     using RenderableFilter = std::function<bool(QSSGModelContext *)>;
 
     QSSGLayerRenderData(QSSGRenderLayer &inLayer, QSSGRenderer &inRenderer);
@@ -210,9 +258,7 @@ public:
                                 QSSGRenderableObjectList &transparentObjects,
                                 QSSGRenderableObjectList &screenTextureObjects,
                                 float lodThreshold = 0.0f);
-    bool prepareParticlesForRender(const RenderableNodeEntries &renderableParticles, const QSSGRenderCameraData &cameraData);
-    bool prepareItem2DsForRender(const QSSGRenderContextInterface &ctxIfc,
-                                 const RenderableItem2DEntries &renderableItem2Ds);
+    bool prepareParticlesForRender(const RenderableNodeEntries &renderableParticles, const QSSGRenderCameraData &cameraData, QSSGLayerRenderPreparationResultFlags &ioFlags);
 
     void prepareResourceLoaders();
 
@@ -236,8 +282,6 @@ public:
 
     void resetForFrame();
 
-    void maybeBakeLightmap();
-
     QSSGFrameData &getFrameData();
 
     ShadowMapPass shadowMapPass;
@@ -245,6 +289,7 @@ public:
     ZPrePassPass zPrePassPass;
     SSAOMapPass ssaoMapPass;
     DepthMapPass depthMapPass;
+    DepthMapPass depthMapPassMS;
     ScreenMapPass screenMapPass;
     ScreenReflectionPass reflectionPass;
     Item2DPass item2DPass;
@@ -254,6 +299,8 @@ public:
     UserPass userPasses[USERPASSES];
     OpaquePass opaquePass;
     TransparentPass transparentPass;
+    OITRenderPass oitRenderPass;
+    OITCompositePass oitCompositePass;
     InfiniteGridPass infiniteGridPass;
     DebugDrawPass debugDrawPass;
 
@@ -265,13 +312,22 @@ public:
     // List of nodes we can render, not all may be active.  Found by doing a depth-first
     // search through m_FirstChild if length is zero.
 
+    using LayerNodes = std::vector<QSSGRenderNode *>;
+    QSSGGlobalRenderNodeData::LayerNodeView layerNodes;
+    LayerNodes layerNodesCategorized;
+
     // renderableNodes have all lights, but properties configured for specific node
     RenderableNodeEntries renderableModels;
     RenderableNodeEntries renderableParticles;
-    QVector<QSSGRenderItem2D *> renderableItem2Ds;
-    QVector<QSSGRenderCamera *> cameras;
-    QVector<QSSGRenderLight *> lights;
-    QVector<QSSGRenderReflectionProbe *> reflectionProbes;
+
+    // Views into the collected nodes (unsorted)
+    QSSGModelsView modelsView;
+    QSSGParticlesView particlesView;
+    QSSGItem2DsView item2DsView;
+    QSSGCamerasView camerasView;
+    QSSGLightsView lightsView;
+    QSSGReflectionProbesView reflectionProbesView;
+    QSSGNonCategorizedView nonCategorizedView;
 
     // Results of prepare for render.
     QSSGRenderCameraList renderedCameras; // multiple items with multiview, one otherwise (or zero if no cameras at all)
@@ -291,14 +347,17 @@ public:
 
     bool tooManyLightsWarningShown = false;
     bool tooManyShadowLightsWarningShown = false;
+    bool oitWarningUnsupportedShown = false;
+    bool oitWarningInvalidBlendModeShown = false;
+    bool orderIndependentTransparencyEnabled = false;
 
-    QSSGLightmapper *m_lightmapper = nullptr;
+    std::unique_ptr<QSSGLightmapBaker> lightmapBaker = nullptr;
 
     QSSGShaderFeatures getShaderFeatures() const { return features; }
     QSSGRhiGraphicsPipelineState getPipelineState() const { return ps; }
 
-    bool interactiveLightmapBakingRequested = false;
-    QSSGLightmapper::Callback lightmapBakingOutputCallback;
+    void initializeLightmapBaking(QSSGLightmapBaker::Context &ctx);
+    void maybeProcessLightmapBaking();
 
     [[nodiscard]] QSSGRenderGraphObject *getCamera(QSSGCameraId id) const;
     [[nodiscard]] QSSGRenderCamera *activeCamera() const { return !renderedCameras.isEmpty() ? renderedCameras[0] : nullptr; }
@@ -339,6 +398,8 @@ public:
     const QSSGRenderShadowMapPtr &getShadowMapManager() const { return shadowMapManager; }
     const QSSGRenderReflectionMapPtr &getReflectionMapManager() const { return reflectionMapManager; }
 
+    QSSGOITRenderContext &getOitRenderContext() { return oitRenderContext; }
+
     static bool prepareInstancing(QSSGRhiContext *rhiCtx,
                                   QSSGSubsetRenderable *renderable,
                                   const QVector3D &cameraDirection,
@@ -350,6 +411,8 @@ public:
     [[nodiscard]] const QSSGRhiRenderableTexture *getRenderResult(QSSGFrameData::RenderResult id) const { return &renderResults[size_t(id)]; }
     [[nodiscard]] static inline const std::unique_ptr<QSSGPerFrameAllocator> &perFrameAllocator(QSSGRenderContextInterface &ctx);
     [[nodiscard]] static inline QSSGLayerRenderData *getCurrent(const QSSGRenderer &renderer) { return renderer.m_currentLayer; }
+    void saveRenderState(const QSSGRenderer &renderer);
+    void restoreRenderState(QSSGRenderer &renderer);
 
     static void setTonemapFeatures(QSSGShaderFeatures &features, QSSGRenderLayer::TonemapMode tonemapMode)
     {
@@ -375,7 +438,7 @@ public:
     QMatrix4x4 getGlobalTransform(QSSGPrepContextId prepId, const QSSGRenderModel &model);
     void setGlobalOpacity(QSSGRenderablesId renderablesId, const QSSGRenderModel &model, float opacity);
     float getGlobalOpacity(QSSGPrepContextId prepId, const QSSGRenderModel &model);
-    [[nodiscard]] QMatrix4x4 getModelMvp(QSSGPrepContextId prepId, const QSSGRenderModel &model) const;
+    [[nodiscard]] QMatrix4x4 getModelMvps(QSSGPrepContextId prepId, const QSSGRenderModel &model) const;
     void setModelMaterials(QSSGRenderablesId renderablesId, const QSSGRenderModel &model, const QList<QSSGResourceId> &materials);
     void setModelMaterials(const QSSGRenderablesId renderablesId, const QList<QSSGResourceId> &materials);
     [[nodiscard]] QSSGPrepResultId prepareModelsForRender(QSSGRenderContextInterface &contextInterface,
@@ -383,6 +446,73 @@ public:
                                                                QSSGRenderablesId renderablesId,
                                                                float lodThreshold);
 
+    // Convenience wrappers for getting values from the node, model store.
+    [[nodiscard]] QMatrix4x4 getGlobalTransform(QSSGRenderNodeHandle h, QMatrix4x4 defaultValue) const
+    {
+        return nodeData->getGlobalTransform(h, defaultValue);
+    }
+    [[nodiscard]] QMatrix4x4 getGlobalTransform(QSSGRenderNodeHandle h) const
+    {
+        return nodeData->getGlobalTransform(h, QMatrix4x4());
+    }
+    [[nodiscard]] QMatrix4x4 getGlobalTransform(const QSSGRenderNode &node) const
+    {
+        return nodeData->getGlobalTransform(node.h, node.localTransform);
+    }
+    [[nodiscard]] QMatrix3x3 getNormalMatrix(QSSGRenderModelHandle h) const
+    {
+        return modelData->getNormalMatrix(h, QMatrix3x3(Qt::Uninitialized));
+    }
+    [[nodiscard]] QMatrix3x3 getNormalMatrix(const QSSGRenderModel &model) const
+    {
+        return modelData->getNormalMatrix(model);
+    }
+    [[nodiscard]] ModelViewProjections getModelMvps(QSSGRenderModelHandle h) const
+    {
+        return modelData->getModelViewProjection(h);
+    }
+    [[nodiscard]] ModelViewProjections getModelMvps(const QSSGRenderModel &model) const
+    {
+        return modelData->getModelViewProjection(model);
+    }
+    [[nodiscard]] InstanceTransforms getInstanceTransforms(QSSGRenderNodeHandle h) const
+    {
+        return nodeData->getInstanceTransforms(h);
+    }
+    [[nodiscard]] InstanceTransforms getInstanceTransforms(const QSSGRenderNode &node) const
+    {
+        return nodeData->getInstanceTransforms(node.h);
+    }
+    [[nodiscard]] float getGlobalOpacity(QSSGRenderNodeHandle h, float defaultValue = 1.0f) const
+    {
+        return nodeData->getGlobalOpacity(h, defaultValue);
+    }
+    [[nodiscard]] float getGlobalOpacity(const QSSGRenderNode &node) const
+    {
+        return nodeData->getGlobalOpacity(node.h);
+    }
+
+    //
+    [[nodiscard]] QSSGRenderItem2DData::Item2DRenderer getItem2DRenderer(const QSSGRenderItem2D &item) const
+    {
+        return item2DData->getItem2DRenderer(item);
+    }
+
+    [[nodiscard]] const std::unique_ptr<QRhiRenderPassDescriptor> &getItem2DRenderPassDescriptor() const
+    {
+        return item2DData->getItem2DRenderPassDescriptor();
+    }
+
+    [[nodiscard]] ModelViewProjections getItem2DMvps(QSSGRenderItem2DHandle h) const
+    {
+        return item2DData->getModelViewProjection(h);
+    }
+
+    [[nodiscard]] ModelViewProjections getItem2DMvps(const QSSGRenderItem2D &item) const
+    {
+        return item2DData->getModelViewProjection(item);
+    }
+
 
     //
     void prepareRenderables(QSSGRenderContextInterface &ctx,
@@ -393,6 +523,12 @@ public:
     void renderRenderables(QSSGRenderContextInterface &ctx,
                            QSSGPrepResultId prepId);
 
+    static bool calculateGlobalVariables(QSSGRenderNode &node,
+                                         std::vector<QMatrix4x4> &globalTransforms,
+                                         std::vector<float> &globalOpacities);
+
+    QSSGRenderCameraData getCameraDataImpl(const QSSGRenderCamera *camera) const;
+
 private:
     friend class QSSGRenderer;
     friend class QSSGRendererPrivate;
@@ -400,8 +536,13 @@ private:
     friend class QSSGModelHelpers;
     friend class QSSGRenderHelpers;
 
-    struct ExtensionContext
+    class ExtensionContext
     {
+    public:
+        explicit ExtensionContext() = default;
+        explicit ExtensionContext(const QSSGRenderExtension &ownerExt, QSSGRenderCamera *cam, size_t idx, quint32 slot)
+            : owner(&ownerExt), camera(cam), ps{}, filter{0}, index(idx), slot(slot)
+        { }
         const QSSGRenderExtension *owner = nullptr;
         QSSGRenderCamera *camera = nullptr;
         QSSGRhiGraphicsPipelineState ps[3] {};
@@ -410,21 +551,25 @@ private:
         quint32 slot = 0;
     };
 
-    std::vector<ExtensionContext> extContexts { { /* 0 - Always available */ } };
-    std::vector<RenderableNodeEntries> renderableModelStore { { /* 0 - Always available */ } };
-    std::vector<TModelContextPtrList> modelContextStore { { /* 0 - Always available */ }};
-    std::vector<QSSGRenderableObjectList> renderableObjectStore { { /* 0 - Always available */ }};
-    std::vector<QSSGRenderableObjectList> opaqueObjectStore { { /* 0 - Always available */ }};
-    std::vector<QSSGRenderableObjectList> transparentObjectStore { { /* 0 - Always available */ }};
-    std::vector<QSSGRenderableObjectList> screenTextureObjectStore { { /* 0 - Always available */ }};
+    std::vector<ExtensionContext> extContexts { ExtensionContext{ /* 0 - Always available */ } };
+    std::vector<RenderableNodeEntries> renderableModelStore { RenderableNodeEntries{ /* 0 - Always available */ } };
+    std::vector<TModelContextPtrList> modelContextStore { TModelContextPtrList{ /* 0 - Always available */ }};
+    std::vector<QSSGRenderableObjectList> renderableObjectStore { QSSGRenderableObjectList{ /* 0 - Always available */ }};
+    std::vector<QSSGRenderableObjectList> opaqueObjectStore { QSSGRenderableObjectList{ /* 0 - Always available */ }};
+    std::vector<QSSGRenderableObjectList> transparentObjectStore { QSSGRenderableObjectList{ /* 0 - Always available */ }};
+    std::vector<QSSGRenderableObjectList> screenTextureObjectStore { QSSGRenderableObjectList{ /* 0 - Always available */ }};
+
+    std::shared_ptr<QSSGGlobalRenderNodeData> nodeData;
+    std::unique_ptr<QSSGRenderModelData> modelData;
+    std::unique_ptr<QSSGRenderItem2DData> item2DData;
 
     // Soreted cache (per camera and extension)
     using PerCameraCache = std::unordered_map<const QSSGRenderCamera *, QSSGRenderableObjectList>;
-    std::vector<PerCameraCache> sortedOpaqueObjectCache { { /* 0 - Always available */ } };
-    std::vector<PerCameraCache> sortedTransparentObjectCache { { /* 0 - Always available */ } };
-    std::vector<PerCameraCache> sortedScreenTextureObjectCache { { /* 0 - Always available */ } };
-    std::vector<PerCameraCache> sortedOpaqueDepthPrepassCache { { /* 0 - Always available */ } };
-    std::vector<PerCameraCache> sortedDepthWriteCache { { /* 0 - Always available */ } };
+    std::vector<PerCameraCache> sortedOpaqueObjectCache { PerCameraCache{ /* 0 - Always available */ } };
+    std::vector<PerCameraCache> sortedTransparentObjectCache { PerCameraCache{ /* 0 - Always available */ } };
+    std::vector<PerCameraCache> sortedScreenTextureObjectCache { PerCameraCache{ /* 0 - Always available */ } };
+    std::vector<PerCameraCache> sortedOpaqueDepthPrepassCache { PerCameraCache{ /* 0 - Always available */ } };
+    std::vector<PerCameraCache> sortedDepthWriteCache { PerCameraCache{ /* 0 - Always available */ } };
 
     [[nodiscard]] const QSSGRenderCameraDataList &getCachedCameraDatas();
     void ensureCachedCameraDatas();
@@ -446,27 +591,37 @@ private:
     static void prepareModelMaterials(RenderableNodeEntries &renderableModels, bool cullUnrenderables);
     static void prepareModelMaterials(const RenderableNodeEntries::ConstIterator &begin,
                                       const RenderableNodeEntries::ConstIterator &end);
-    // Load meshes as needed
-    static void prepareModelMeshes(const QSSGRenderContextInterface &contextInterface,
-                                   RenderableNodeEntries &renderableModels,
-                                   bool globalPickingEnabled);
-    static void prepareModelMeshes(const QSSGRenderContextInterface &contextInterface,
-                                   const RenderableNodeEntries::ConstIterator begin,
-                                   const RenderableNodeEntries::ConstIterator end,
-                                   bool globalPickingEnabled);
 
     // Persistent data
     QHash<QSSGShaderMapKey, QSSGRhiShaderPipelinePtr> shaderMap;
 
+    // Cached buffer.
+    QByteArray generatedShaderString;
+
+    // Saved render state (for sublayers)
+    struct SavedRenderState
+    {
+        QRect viewport;
+        QRect scissorRect;
+        float dpr = 1.0;
+    };
+
+    std::optional<SavedRenderState> savedRenderState;
+
     // Note: Re-used to avoid expensive initialization.
     // - Should be revisit, as we can do better.
     QSSGShaderDefaultMaterialKeyProperties defaultMaterialShaderKeyProperties;
     QSSGFrameData frameData;
     QSSGRhiGraphicsPipelineState ps; // Base pipleline state
     QSSGShaderFeatures features; // Base feature set
+    quint32 version = 0;
     bool particlesEnabled = true;
     bool hasDepthWriteObjects = false;
     bool zPrePassActive = false;
+    // NOTE: For the time being we need to keep track of extensions modifying the renderables
+    // because then we need to reset the lists.
+    // FIXME: This should be revisited, as we can do better (hence the verbose name).
+    bool renderablesModifiedByExtension = false;
     enum class DepthPrepassObject : quint8
     {
         None = 0x0,
@@ -480,7 +635,8 @@ private:
     QSSGRenderReflectionMapPtr reflectionMapManager;
     QHash<const QSSGModelContext *, QRhiTexture *> lightmapTextures;
     QHash<const QSSGModelContext *, QRhiTexture *> bonemapTextures;
-    QSSGRhiRenderableTexture renderResults[3] {};
+    QSSGRhiRenderableTexture renderResults[6] {};
+    QSSGOITRenderContext oitRenderContext;
 };
 
 QT_END_NAMESPACE
